package com.telpo.dipperposition.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.telpo.dipperposition.common.*;
import com.telpo.dipperposition.config.PositionConfig;
import com.telpo.dipperposition.entity.mongo.IpProvinceEntity;
import com.telpo.dipperposition.entity.mongo.ProvinceInfoEntity;
import com.telpo.dipperposition.service.IDipperAstPosAsyncTaskService;
import com.telpo.dipperposition.service.IProvinceInfoService;
import com.telpo.dipperposition.service.IpProvinceService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ObjectUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.HashMap;
import java.util.Map;

/**
 * @program: DipperAstPosAsyncTaskServiceImpl
 * @description: 系统预先基于省份的省会城市的经纬度作为辅助信息，
 *  * 根据设备请求的IP地址，从高德IP定位服务获取相关的省份，再匹配相应的位置信息作为辅助位置信息。
 *  * 如果匹配不到省份，则以武汉中心作为辅助为位置信息。
 *  * 关于IP与省份的关系保存到缓存中，用于下次使用时，先在缓存中获取匹配信息，匹配不到，再请求高德IP定位服务。
 *  * 高德IP定位服务：https://lbs.amap.com/api/webservice/guide/api/ipconfig。
 * @author: king
 * @create: 2021-01-10 14:01
 */
@Service
@Slf4j
public class DipperAstPosAsyncTaskServiceImpl implements IDipperAstPosAsyncTaskService {

    @Autowired
    private OkHttpUtil okHttpUtil;
    @Autowired
    private PositionConfig positionConfig;
    @Autowired
    private IpProvinceService ipProvinceService;
    @Autowired
    private IProvinceInfoService provinceInfoService;

    private String centerProvince;
    private String ipPositionRequestPath;
    private String ipPositionRequestKey;

    // 根据IP获取省会信息
    private String getIpPositionProvince(String ipAddress) {

        // 关于IP与省份的关系保存到缓存中
        // 使用时，先在缓存中获取匹配信息
        // 用mongodb实现
        IpProvinceEntity ipProvinceEntity = ipProvinceService.getIpProvince(ipAddress);
        if (ipProvinceEntity == null) {
            // 匹配不到，再请求高德IP定位服务。
            //JSONObject userObj = new JSONObject();
            Map<String, String> dataMap = new HashMap<String, String>();
            dataMap.put("ip", ipAddress);
            dataMap.put("key", ipPositionRequestKey);
            StringBuffer buffer = okHttpUtil.getQueryString(ipPositionRequestPath, dataMap);
            JSONObject paramObject = new JSONObject();
            String  result  =  okHttpUtil.getJsonParams(buffer.toString(), paramObject.toJSONString());
            if (ObjectUtils.isNotEmpty(result)) {
                if (ObjectUtils.isEmpty(result)) {
                    log.debug("IP省份获取错误，结果为空");
                    return null;
                } else {
                    if (result.equals("FAIL")) {
                        log.debug("IP省份获取错误，结果为FAIL");
                        return null;
                    }
                    //log.debug("IP省份获取错误，结果为：" + result);
                    JSONObject provinceJson = JSONObject.parseObject(result);
                    String province = null;
                    try {
                        province = (String)provinceJson.get("province");
                        if (ObjectUtils.isEmpty(province)) {
                            log.debug("json is :" + buffer.toString());
                            return null;
                        }
                    } catch (Exception e) {
                        log.error("JSONObject分析出错，provinceJson：" + provinceJson);
                    }

                    return province;
                }
            } else {
                // 意外错误
                log.debug("ip address is null");
                return null;
            }
        } else {
            return ipProvinceEntity.getProvince();
        }
    }

    // 将IP对应的省会保存到mongoDB
    public void createIPProvince(String ipAddress, String province) {
        log.debug("创建IP省份记录!");
        try {
            IpProvinceEntity ipProvinceEntity = ipProvinceService.getIpProvince(ipAddress);
            if (ipProvinceEntity == null) {
                ipProvinceEntity = new IpProvinceEntity();
                ipProvinceEntity.setIp(ipAddress);
                ipProvinceEntity.setProvince(province);
                ipProvinceService.saveIpProvince(ipProvinceEntity);
            }
        } catch (Exception e) {
            log.error("创建IP省份记录异常:", e);
        }
    }

    /*
     * 获取定位辅助信息
     * @param ipAddress
     */
    @Override
    //@Async("asyncServiceExecutor")
    public String pushAstPos(String ipAddress) {
        this.ipPositionRequestKey = positionConfig.getIpPositionRequestKey();
        this.ipPositionRequestPath = positionConfig.getIpPositionRequestPath();
        this.centerProvince = positionConfig.getCenterProvince();
        //this.centerProvinceFilePath = positionConfig.getCenterProvinceFilePath();
        // (1) 获取省会城市信息
        String centerAddress = getIpPositionProvince(ipAddress);
        if (centerAddress == null || ObjectUtils.isEmpty(centerAddress) || centerAddress.equals("0")) {
            log.warn("IP地址非法，无法获取辅助位置信息！");
            // 返回武汉的定位数据
            centerAddress = this.centerProvince;
        } else {
            // 保存到mongoDB
            createIPProvince(ipAddress, centerAddress);
        }

        String lonValue = null;
        String altValue = null;
        try {
            byte[] utf8 = centerAddress.getBytes("UTF-8");
            String utf8CenterAddress = new String(utf8, "UTF-8");
            log.debug("centerAddress is " + utf8CenterAddress);

            if (utf8CenterAddress.endsWith("省") || utf8CenterAddress.endsWith("区") || utf8CenterAddress.endsWith("市")) {

                ProvinceInfoEntity entity = provinceInfoService.getProvinceInfo(utf8CenterAddress);
                if (entity != null) {
                    lonValue = entity.getLon();
                    altValue = entity.getAlt();
                }
            } else {
                log.error("非法省份：" + utf8CenterAddress);
                return null;
            }
        } catch (Exception e) {
            log.error(e.getMessage());
        }


        // （2） 处理返回结果
        if (lonValue == null) {
            // null处理
            log.error("系统错误,请联系系统管理员。");
            return null;
        }

        // push to GNNS Server
        return getCmdOfPos(lonValue, altValue);
    }

    // 组装命令发送给设备
    private String getCmdOfPos(String lonStr, String altStr) {

        // 时间和位置不是从服务器获取，而是本地生成
//        String[] astPosArray = astPos.split(",");
//        String lan = astPosArray[0].trim();
//        String alt = astPosArray[1].trim();
        double lanValue = Double.parseDouble(lonStr) * 10000000;
        long lanLongValue = Double.doubleToLongBits(lanValue);
        if (lanLongValue < 0) {
            // FFFFFFFF - 439C3270 + 1= BC63CD90（补码）
            lanLongValue = lanLongValue + 4294967295L + 1;
        }
        double altValue = Double.parseDouble(altStr) * 10000000;
        long altLongValue = Double.doubleToLongBits(altValue);
        if (altLongValue < 0) {
            // FFFFFFFF - 439C3270 + 1= BC63CD90（补码）
            altLongValue = altLongValue + 4294967295L + 1;
        }

        // 数值换算举例（以经度举例。纬度、高度、位置精度换算方法一致）：
        // （1）经度数值为 113.431，则换算方法如下：
        //     113.431/比例因子 = 1134310000（十进制）
        //      439C3270（十六进制）
        //      经度数据填入 70 32 9C 43（小端模式）
        // （2）经度数值为-113.431，则换算方法如下：
        //     113.431/比例因子 = 1134310000（十进制）
        //      439C3270（十六进制）
        //      FFFFFFFF - 439C3270 + 1= BC63CD90（补码）
        //  经度数据填入 90 CD 63 BC（小端模式）
        //  指令（十六进制）
        //  举例： 23 3E 04 01 10 00 70 32 9C 43 D0 B2 CE 0D 70 17 00 00 40 0D 03 00 CA 95
        //  其中
        //  23 3E 为同步头
        //  04 01 为识别码
        //  10 00 表示长度为 16
        //  70 32 9C 43 表示注入的辅助经度为 113.431 度
        //  D0 B2 CE 0D 表示注入的辅助纬度为 23.165 度
        //  70 17 00 00 表示注入的辅助高度为 60 米
        //  40 0D 03 00 表示注入的位置精度为 2000 米 030d40
        //  A0 86 01 00 表示注入的位置精度为 1000 米 0186a0
        //  00 2F 为校验和

        // astPosCmd 组装
        StringBuilder astPosCmdBuf = new StringBuilder();
        astPosCmdBuf.append("233E0401");
        astPosCmdBuf.append("1000");


        // 小端模式
        String lanString = HexConvert.encodeHEX(lanLongValue).toUpperCase();
        String altSting = HexConvert.encodeHEX(altLongValue).toUpperCase();
        //astPosCmdBuf.append(HexConvert.encodeHEX(lanLongValue).toUpperCase());
        astPosCmdBuf.append(lanString.substring(6,8));
        astPosCmdBuf.append(lanString.substring(4,6));
        astPosCmdBuf.append(lanString.substring(2,4));
        astPosCmdBuf.append(lanString.substring(0,2));

        //astPosCmdBuf.append(HexConvert.encodeHEX(altLongValue).toUpperCase());

        astPosCmdBuf.append(altSting.substring(6,8));
        astPosCmdBuf.append(altSting.substring(4,6));
        astPosCmdBuf.append(altSting.substring(2,4));
        astPosCmdBuf.append(altSting.substring(0,2));
        astPosCmdBuf.append("70170000");
        astPosCmdBuf.append("A0860100");

        //log.info(astPosCmd);
//        String checkSum = HexConvert.makeChecksum(astPosCmdBuf.toString()).toUpperCase();
//        StringBuffer astCheckSumBuf = new StringBuffer();
//        astCheckSumBuf.append(checkSum);
//        while (astCheckSumBuf.length()<4) {
//            astCheckSumBuf.insert(0,"0");
//        }
//        checkSum = astCheckSumBuf.toString();
//        //log.info(checkSum);
//
//        byte[] astPosCmdBytes = HexConvert.hexStringToBytes(astPosCmdBuf.toString());
//        StringBuilder astPosCmdNewBuf = new StringBuilder();
//        for (byte astPosCmdByte : astPosCmdBytes) {
//            String s = Integer.toHexString(astPosCmdByte & 0xff);
//            if (s.length() < 2) {
//                astPosCmdNewBuf.append('0');
//            }
//            astPosCmdNewBuf.append(s + " ");
//        }
//
        return astPosCmdBuf.toString();
        //+ checkSum.substring(0,2) + " " + checkSum.substring(2,4);



    }
}
